@@ -1,11 +1,11 @@
 //! `cost_tracker` keeps tracking transaction cost per chained accounts as well as for entire block
 //! The main functions are:
-//! - would_transaction_fit(&tx_cost), immutable function to test if tx with tx_cost would fit into current block
+//! - would_fit(&tx_cost), immutable function to test if tx with tx_cost would fit into current block
 //! - add_transaction_cost(&tx_cost), mutable function to accumulate tx_cost to tracker.
 //!
 use {
     crate::{block_cost_limits::*, cost_model::TransactionCost},
-    solana_sdk::{clock::Slot, pubkey::Pubkey, transaction::SanitizedTransaction},
+    solana_sdk::{clock::Slot, pubkey::Pubkey},
     std::collections::HashMap,
 };
 
@@ -36,7 +36,21 @@ pub struct CostTracker {
 
 impl Default for CostTracker {
     fn default() -> Self {
-        CostTracker::new(MAX_WRITABLE_ACCOUNT_UNITS, MAX_BLOCK_UNITS, MAX_VOTE_UNITS)
+        // Clippy doesn't like asserts in const contexts, so need to explicitly allow them.  For
+        // more info, see this issue: https://github.com/rust-lang/rust-clippy/issues/8159
+        #![allow(clippy::assertions_on_constants)]
+        const _: () = assert!(MAX_WRITABLE_ACCOUNT_UNITS <= MAX_BLOCK_UNITS);
+        const _: () = assert!(MAX_VOTE_UNITS <= MAX_BLOCK_UNITS);
+
+        Self {
+            account_cost_limit: MAX_WRITABLE_ACCOUNT_UNITS,
+            block_cost_limit: MAX_BLOCK_UNITS,
+            vote_cost_limit: MAX_VOTE_UNITS,
+            cost_by_writable_accounts: HashMap::with_capacity(WRITABLE_ACCOUNTS_PER_BLOCK),
+            block_cost: 0,
+            vote_cost: 0,
+            transaction_count: 0,
+        }
     }
 }
 
@@ -67,33 +81,32 @@ impl CostTracker {
         self.vote_cost_limit = vote_cost_limit;
     }
 
-    pub fn would_transaction_fit(
-        &self,
-        transaction: &SanitizedTransaction,
-        tx_cost: &TransactionCost,
-    ) -> Result<(), CostTrackerError> {
-        self.would_fit(&tx_cost.writable_accounts, tx_cost.sum(), transaction)
-    }
-
-    pub fn add_transaction_cost(
-        &mut self,
-        transaction: &SanitizedTransaction,
-        tx_cost: &TransactionCost,
-    ) {
-        self.add_transaction(&tx_cost.writable_accounts, tx_cost.sum(), transaction);
-    }
-
-    pub fn try_add(
-        &mut self,
-        transaction: &SanitizedTransaction,
-        tx_cost: &TransactionCost,
-    ) -> Result<u64, CostTrackerError> {
-        let cost = tx_cost.sum();
-        self.would_fit(&tx_cost.writable_accounts, cost, transaction)?;
-        self.add_transaction(&tx_cost.writable_accounts, cost, transaction);
+    pub fn try_add(&mut self, tx_cost: &TransactionCost) -> Result<u64, CostTrackerError> {
+        self.would_fit(tx_cost)?;
+        self.add_transaction_cost(tx_cost);
         Ok(self.block_cost)
     }
 
+    pub fn update_execution_cost(
+        &mut self,
+        _estimated_tx_cost: &TransactionCost,
+        _actual_execution_cost: u64,
+    ) {
+        // TODO: adjust block_cost / vote_cost / account_cost by (actual_execution_cost - execution_cost)
+    }
+
+    pub fn remove(&mut self, tx_cost: &TransactionCost) {
+        self.remove_transaction_cost(tx_cost);
+    }
+
+    pub fn block_cost(&self) -> u64 {
+        self.block_cost
+    }
+
+    pub fn transaction_count(&self) -> u64 {
+        self.transaction_count
+    }
+
     pub fn report_stats(&self, bank_slot: Slot) {
         // skip reporting if block is empty
         if self.transaction_count == 0 {
@@ -131,21 +144,18 @@ impl CostTracker {
         (costliest_account, costliest_account_cost)
     }
 
-    fn would_fit(
-        &self,
-        keys: &[Pubkey],
-        cost: u64,
-        transaction: &SanitizedTransaction,
-    ) -> Result<(), CostTrackerError> {
+    fn would_fit(&self, tx_cost: &TransactionCost) -> Result<(), CostTrackerError> {
+        let writable_accounts = &tx_cost.writable_accounts;
+        let cost = tx_cost.sum();
+        let vote_cost = if tx_cost.is_simple_vote { cost } else { 0 };
+
         // check against the total package cost
         if self.block_cost.saturating_add(cost) > self.block_cost_limit {
             return Err(CostTrackerError::WouldExceedBlockMaxLimit);
         }
 
         // if vote transaction, check if it exceeds vote_transaction_limit
-        if transaction.is_simple_vote_transaction()
-            && self.vote_cost.saturating_add(cost) > self.vote_cost_limit
-        {
+        if self.vote_cost.saturating_add(vote_cost) > self.vote_cost_limit {
             return Err(CostTrackerError::WouldExceedVoteMaxLimit);
         }
 
@@ -155,7 +165,7 @@ impl CostTracker {
         }
 
         // check each account against account_cost_limit,
-        for account_key in keys.iter() {
+        for account_key in writable_accounts.iter() {
             match self.cost_by_writable_accounts.get(account_key) {
                 Some(chained_cost) => {
                     if chained_cost.saturating_add(cost) > self.account_cost_limit {
@@ -171,8 +181,9 @@ impl CostTracker {
         Ok(())
     }
 
-    fn add_transaction(&mut self, keys: &[Pubkey], cost: u64, transaction: &SanitizedTransaction) {
-        for account_key in keys.iter() {
+    fn add_transaction_cost(&mut self, tx_cost: &TransactionCost) {
+        let cost = tx_cost.sum();
+        for account_key in tx_cost.writable_accounts.iter() {
             let account_cost = self
                 .cost_by_writable_accounts
                 .entry(*account_key)
@@ -180,11 +191,27 @@ impl CostTracker {
             *account_cost = account_cost.saturating_add(cost);
         }
         self.block_cost = self.block_cost.saturating_add(cost);
-        if transaction.is_simple_vote_transaction() {
+        if tx_cost.is_simple_vote {
             self.vote_cost = self.vote_cost.saturating_add(cost);
         }
         self.transaction_count = self.transaction_count.saturating_add(1);
     }
+
+    fn remove_transaction_cost(&mut self, tx_cost: &TransactionCost) {
+        let cost = tx_cost.sum();
+        for account_key in tx_cost.writable_accounts.iter() {
+            let account_cost = self
+                .cost_by_writable_accounts
+                .entry(*account_key)
+                .or_insert(0);
+            *account_cost = account_cost.saturating_sub(cost);
+        }
+        self.block_cost = self.block_cost.saturating_sub(cost);
+        if tx_cost.is_simple_vote {
+            self.vote_cost = self.vote_cost.saturating_sub(cost);
+        }
+        self.transaction_count = self.transaction_count.saturating_sub(1);
+    }
 }
 
 #[cfg(test)]
@@ -199,7 +226,7 @@ mod tests {
             hash::Hash,
             signature::{Keypair, Signer},
             system_transaction,
-            transaction::{TransactionError, VersionedTransaction},
+            transaction::{SanitizedTransaction, TransactionError, VersionedTransaction},
         },
         solana_vote_program::vote_transaction,
         std::{cmp, sync::Arc},
@@ -220,19 +247,22 @@ mod tests {
     fn build_simple_transaction(
         mint_keypair: &Keypair,
         start_hash: &Hash,
-    ) -> (SanitizedTransaction, Vec<Pubkey>, u64) {
+    ) -> (SanitizedTransaction, TransactionCost) {
         let keypair = Keypair::new();
         let simple_transaction = SanitizedTransaction::from_transaction_for_tests(
             system_transaction::transfer(mint_keypair, &keypair.pubkey(), 2, *start_hash),
         );
+        let mut tx_cost = TransactionCost::new_with_capacity(1);
+        tx_cost.execution_cost = 5;
+        tx_cost.writable_accounts.push(mint_keypair.pubkey());
 
-        (simple_transaction, vec![mint_keypair.pubkey()], 5)
+        (simple_transaction, tx_cost)
     }
 
     fn build_simple_vote_transaction(
         mint_keypair: &Keypair,
         start_hash: &Hash,
-    ) -> (SanitizedTransaction, Vec<Pubkey>, u64) {
+    ) -> (SanitizedTransaction, TransactionCost) {
         let keypair = Keypair::new();
         let transaction = vote_transaction::new_vote_transaction(
             vec![42],
@@ -251,7 +281,12 @@ mod tests {
             |_| Err(TransactionError::UnsupportedVersion),
         )
         .unwrap();
-        (vote_transaction, vec![mint_keypair.pubkey()], 10)
+        let mut tx_cost = TransactionCost::new_with_capacity(1);
+        tx_cost.execution_cost = 10;
+        tx_cost.writable_accounts.push(mint_keypair.pubkey());
+        tx_cost.is_simple_vote = true;
+
+        (vote_transaction, tx_cost)
     }
 
     #[test]
@@ -267,12 +302,13 @@ mod tests {
     #[test]
     fn test_cost_tracker_ok_add_one() {
         let (mint_keypair, start_hash) = test_setup();
-        let (tx, keys, cost) = build_simple_transaction(&mint_keypair, &start_hash);
+        let (_tx, tx_cost) = build_simple_transaction(&mint_keypair, &start_hash);
+        let cost = tx_cost.sum();
 
         // build testee to have capacity for one simple transaction
         let mut testee = CostTracker::new(cost, cost, cost);
-        assert!(testee.would_fit(&keys, cost, &tx).is_ok());
-        testee.add_transaction(&keys, cost, &tx);
+        assert!(testee.would_fit(&tx_cost).is_ok());
+        testee.add_transaction_cost(&tx_cost);
         assert_eq!(cost, testee.block_cost);
         assert_eq!(0, testee.vote_cost);
         let (_costliest_account, costliest_account_cost) = testee.find_costliest_account();
@@ -282,12 +318,13 @@ mod tests {
     #[test]
     fn test_cost_tracker_ok_add_one_vote() {
         let (mint_keypair, start_hash) = test_setup();
-        let (tx, keys, cost) = build_simple_vote_transaction(&mint_keypair, &start_hash);
+        let (_tx, tx_cost) = build_simple_vote_transaction(&mint_keypair, &start_hash);
+        let cost = tx_cost.sum();
 
         // build testee to have capacity for one simple transaction
         let mut testee = CostTracker::new(cost, cost, cost);
-        assert!(testee.would_fit(&keys, cost, &tx).is_ok());
-        testee.add_transaction(&keys, cost, &tx);
+        assert!(testee.would_fit(&tx_cost).is_ok());
+        testee.add_transaction_cost(&tx_cost);
         assert_eq!(cost, testee.block_cost);
         assert_eq!(cost, testee.vote_cost);
         let (_ccostliest_account, costliest_account_cost) = testee.find_costliest_account();
@@ -298,18 +335,20 @@ mod tests {
     fn test_cost_tracker_ok_add_two_same_accounts() {
         let (mint_keypair, start_hash) = test_setup();
         // build two transactions with same signed account
-        let (tx1, keys1, cost1) = build_simple_transaction(&mint_keypair, &start_hash);
-        let (tx2, keys2, cost2) = build_simple_transaction(&mint_keypair, &start_hash);
+        let (_tx1, tx_cost1) = build_simple_transaction(&mint_keypair, &start_hash);
+        let cost1 = tx_cost1.sum();
+        let (_tx2, tx_cost2) = build_simple_transaction(&mint_keypair, &start_hash);
+        let cost2 = tx_cost2.sum();
 
         // build testee to have capacity for two simple transactions, with same accounts
         let mut testee = CostTracker::new(cost1 + cost2, cost1 + cost2, cost1 + cost2);
         {
-            assert!(testee.would_fit(&keys1, cost1, &tx1).is_ok());
-            testee.add_transaction(&keys1, cost1, &tx1);
+            assert!(testee.would_fit(&tx_cost1).is_ok());
+            testee.add_transaction_cost(&tx_cost1);
         }
         {
-            assert!(testee.would_fit(&keys2, cost2, &tx2).is_ok());
-            testee.add_transaction(&keys2, cost2, &tx2);
+            assert!(testee.would_fit(&tx_cost2).is_ok());
+            testee.add_transaction_cost(&tx_cost2);
         }
         assert_eq!(cost1 + cost2, testee.block_cost);
         assert_eq!(1, testee.cost_by_writable_accounts.len());
@@ -321,19 +360,21 @@ mod tests {
     fn test_cost_tracker_ok_add_two_diff_accounts() {
         let (mint_keypair, start_hash) = test_setup();
         // build two transactions with diff accounts
-        let (tx1, keys1, cost1) = build_simple_transaction(&mint_keypair, &start_hash);
         let second_account = Keypair::new();
-        let (tx2, keys2, cost2) = build_simple_transaction(&second_account, &start_hash);
+        let (_tx1, tx_cost1) = build_simple_transaction(&mint_keypair, &start_hash);
+        let cost1 = tx_cost1.sum();
+        let (_tx2, tx_cost2) = build_simple_transaction(&second_account, &start_hash);
+        let cost2 = tx_cost2.sum();
 
         // build testee to have capacity for two simple transactions, with same accounts
         let mut testee = CostTracker::new(cmp::max(cost1, cost2), cost1 + cost2, cost1 + cost2);
         {
-            assert!(testee.would_fit(&keys1, cost1, &tx1).is_ok());
-            testee.add_transaction(&keys1, cost1, &tx1);
+            assert!(testee.would_fit(&tx_cost1).is_ok());
+            testee.add_transaction_cost(&tx_cost1);
         }
         {
-            assert!(testee.would_fit(&keys2, cost2, &tx2).is_ok());
-            testee.add_transaction(&keys2, cost2, &tx2);
+            assert!(testee.would_fit(&tx_cost2).is_ok());
+            testee.add_transaction_cost(&tx_cost2);
         }
         assert_eq!(cost1 + cost2, testee.block_cost);
         assert_eq!(2, testee.cost_by_writable_accounts.len());
@@ -345,19 +386,21 @@ mod tests {
     fn test_cost_tracker_chain_reach_limit() {
         let (mint_keypair, start_hash) = test_setup();
         // build two transactions with same signed account
-        let (tx1, keys1, cost1) = build_simple_transaction(&mint_keypair, &start_hash);
-        let (tx2, keys2, cost2) = build_simple_transaction(&mint_keypair, &start_hash);
+        let (_tx1, tx_cost1) = build_simple_transaction(&mint_keypair, &start_hash);
+        let cost1 = tx_cost1.sum();
+        let (_tx2, tx_cost2) = build_simple_transaction(&mint_keypair, &start_hash);
+        let cost2 = tx_cost2.sum();
 
         // build testee to have capacity for two simple transactions, but not for same accounts
         let mut testee = CostTracker::new(cmp::min(cost1, cost2), cost1 + cost2, cost1 + cost2);
         // should have room for first transaction
         {
-            assert!(testee.would_fit(&keys1, cost1, &tx1).is_ok());
-            testee.add_transaction(&keys1, cost1, &tx1);
+            assert!(testee.would_fit(&tx_cost1).is_ok());
+            testee.add_transaction_cost(&tx_cost1);
         }
         // but no more sapce on the same chain (same signer account)
         {
-            assert!(testee.would_fit(&keys2, cost2, &tx2).is_err());
+            assert!(testee.would_fit(&tx_cost2).is_err());
         }
     }
 
@@ -365,21 +408,23 @@ mod tests {
     fn test_cost_tracker_reach_limit() {
         let (mint_keypair, start_hash) = test_setup();
         // build two transactions with diff accounts
-        let (tx1, keys1, cost1) = build_simple_transaction(&mint_keypair, &start_hash);
         let second_account = Keypair::new();
-        let (tx2, keys2, cost2) = build_simple_transaction(&second_account, &start_hash);
+        let (_tx1, tx_cost1) = build_simple_transaction(&mint_keypair, &start_hash);
+        let cost1 = tx_cost1.sum();
+        let (_tx2, tx_cost2) = build_simple_transaction(&second_account, &start_hash);
+        let cost2 = tx_cost2.sum();
 
         // build testee to have capacity for each chain, but not enough room for both transactions
         let mut testee =
             CostTracker::new(cmp::max(cost1, cost2), cost1 + cost2 - 1, cost1 + cost2 - 1);
         // should have room for first transaction
         {
-            assert!(testee.would_fit(&keys1, cost1, &tx1).is_ok());
-            testee.add_transaction(&keys1, cost1, &tx1);
+            assert!(testee.would_fit(&tx_cost1).is_ok());
+            testee.add_transaction_cost(&tx_cost1);
         }
         // but no more room for package as whole
         {
-            assert!(testee.would_fit(&keys2, cost2, &tx2).is_err());
+            assert!(testee.would_fit(&tx_cost2).is_err());
         }
     }
 
@@ -387,26 +432,28 @@ mod tests {
     fn test_cost_tracker_reach_vote_limit() {
         let (mint_keypair, start_hash) = test_setup();
         // build two mocking vote transactions with diff accounts
-        let (tx1, keys1, cost1) = build_simple_vote_transaction(&mint_keypair, &start_hash);
         let second_account = Keypair::new();
-        let (tx2, keys2, cost2) = build_simple_vote_transaction(&second_account, &start_hash);
+        let (_tx1, tx_cost1) = build_simple_vote_transaction(&mint_keypair, &start_hash);
+        let (_tx2, tx_cost2) = build_simple_vote_transaction(&second_account, &start_hash);
+        let cost1 = tx_cost1.sum();
+        let cost2 = tx_cost2.sum();
 
         // build testee to have capacity for each chain, but not enough room for both votes
         let mut testee = CostTracker::new(cmp::max(cost1, cost2), cost1 + cost2, cost1 + cost2 - 1);
         // should have room for first vote
         {
-            assert!(testee.would_fit(&keys1, cost1, &tx1).is_ok());
-            testee.add_transaction(&keys1, cost1, &tx1);
+            assert!(testee.would_fit(&tx_cost1).is_ok());
+            testee.add_transaction_cost(&tx_cost1);
         }
         // but no more room for package as whole
         {
-            assert!(testee.would_fit(&keys2, cost2, &tx2).is_err());
+            assert!(testee.would_fit(&tx_cost2).is_err());
         }
         // however there is room for none-vote tx3
         {
             let third_account = Keypair::new();
-            let (tx3, keys3, cost3) = build_simple_transaction(&third_account, &start_hash);
-            assert!(testee.would_fit(&keys3, cost3, &tx3).is_ok());
+            let (_tx3, tx_cost3) = build_simple_transaction(&third_account, &start_hash);
+            assert!(testee.would_fit(&tx_cost3).is_ok());
         }
     }
 
@@ -414,7 +461,7 @@ mod tests {
     fn test_cost_tracker_try_add_is_atomic() {
         let (mint_keypair, start_hash) = test_setup();
         // build two mocking vote transactions with diff accounts
-        let (tx1, _keys1, _cost1) = build_simple_vote_transaction(&mint_keypair, &start_hash);
+        let (_tx1, _tx_cost1) = build_simple_vote_transaction(&mint_keypair, &start_hash);
 
         let acct1 = Pubkey::new_unique();
         let acct2 = Pubkey::new_unique();
@@ -436,7 +483,7 @@ mod tests {
                 execution_cost: cost,
                 ..TransactionCost::default()
             };
-            assert!(testee.try_add(&tx1, &tx_cost).is_ok());
+            assert!(testee.try_add(&tx_cost).is_ok());
             let (_costliest_account, costliest_account_cost) = testee.find_costliest_account();
             assert_eq!(cost, testee.block_cost);
             assert_eq!(3, testee.cost_by_writable_accounts.len());
@@ -454,7 +501,7 @@ mod tests {
                 execution_cost: cost,
                 ..TransactionCost::default()
             };
-            assert!(testee.try_add(&tx1, &tx_cost).is_ok());
+            assert!(testee.try_add(&tx_cost).is_ok());
             let (costliest_account, costliest_account_cost) = testee.find_costliest_account();
             assert_eq!(cost * 2, testee.block_cost);
             assert_eq!(3, testee.cost_by_writable_accounts.len());
@@ -474,7 +521,7 @@ mod tests {
                 execution_cost: cost,
                 ..TransactionCost::default()
             };
-            assert!(testee.try_add(&tx1, &tx_cost).is_err());
+            assert!(testee.try_add(&tx_cost).is_err());
             let (costliest_account, costliest_account_cost) = testee.find_costliest_account();
             assert_eq!(cost * 2, testee.block_cost);
             assert_eq!(3, testee.cost_by_writable_accounts.len());
